import sys
import unittest
import base64
import mimetypes
import socket
import io
from hashlib import md5
from os import path, urandom
from urllib.parse import urlencode

import httplib2
from wsgi_intercept import WSGIAppError
from wsgi_intercept.interceptor import Httplib2Interceptor

from nanohttp import Controller, configure, Application

TEST_DIR = path.abspath(path.dirname(__file__))
STUFF_DIR = path.join(TEST_DIR, 'stuff')


def md5sum(f):
    if isinstance(f, str):
        file_obj = open(f, 'rb')
    else:
        file_obj = f

    try:
        checksum = md5()
        while True:
            d = file_obj.read(0x8000)
            if not d:
                break
            checksum.update(d)
        return checksum.digest()
    finally:
        if file_obj is not f:
            file_obj.close()


class WsgiTester(httplib2.Http):

    def __init__(self, app_factory, host='nanohttp.org', port=80, **kw):
        super(WsgiTester, self).__init__()
        self.force_exception_to_status_code
        self.interceptor = Httplib2Interceptor(app_factory, host=host, port=port, **kw)

    def __enter__(self):
        self.interceptor.__enter__()
        return self

    def __exit__(self, exc_type, value, traceback):
        self.interceptor.__exit__(exc_type, value, traceback)

    def request(self, uri, query_string=None, fields=None, files=None, json=None, **kw):
        headers = kw.setdefault('headers', {})
        body = None

        if files:
            content_type, body, length = encode_multipart_data(fields, files)
            body = body.getvalue()
            headers.setdefault('Content-Type', content_type)
        elif fields:
            body = urlencode(fields)
            headers.setdefault('Content-Type', 'application/x-www-form-urlencoded')
        elif json:
            body = json
            headers.setdefault('Content-Type', 'application/json')

        if query_string:
            uri += '%s%s' % ('&' if '?' in uri else '?', urlencode(query_string))

        try:
            return super(WsgiTester, self).request(
                '%s%s' % (self.interceptor.url, uri),
                body=body,
                **kw
            )
        except WSGIAppError as e:
            content = b"Internal server error."
            response = httplib2.Response({
                "content-type": "text/plain",
                "status": "408",
                "content-length": len(content)
            })
            response.reason = content
            response.status = 500
            return response, content


class WsgiAppTestCase(unittest.TestCase):

    class Root(Controller):
        pass

    class Application(Application):
        pass

    def setUp(self):
        configure(force=True)
        self.application = self.Application(self.Root())
        self.client = WsgiTester(lambda: self.application)
        self.client.__enter__()

    def tearDown(self):
        self.client.__exit__(*sys.exc_info())

    def assert_request(self, uri, method, cookies=None, expected_response=None, expected_checksum=None,
                       not_expected_headers=None, expected_headers=None, status=200, **kw):

        if cookies:
            headers = kw.setdefault('headers', {})
            if isinstance(cookies, str):
                headers['Cookie'] = cookies
            else:
                headers['Cookie'] = '; '.join('%s=%s' % (k, v.value) for k, v in cookies.items())

        response, content = self.client.request(uri, method=method, **kw)

        # if response.status == 500:
        #     print(content.decode(), file=sys.stderr)
        self.assertEqual(response.status, status)

        if expected_response is not None:
            if isinstance(expected_response, str):
                self.assertEqual(expected_response, content.decode())
            else:
                self.assertRegex(content.decode(), expected_response)

        for k, v in (expected_headers or {}).items():
            k = k.lower()
            if isinstance(v, str):
                self.assertEqual(response[k], v)
            else:
                self.assertRegex(response[k], v)

        for k in not_expected_headers or []:
            self.assertNotIn(k.lower(), response)

        if expected_checksum:
            self.assertEqual(md5sum(io.BytesIO(content)), expected_checksum)

        return response, content

    def assert_get(self, uri, *args, **kw):
        return self.assert_request(uri, 'get', *args, **kw)

    def assert_options(self, uri, *args, **kw):
        return self.assert_request(uri, 'options', *args, **kw)

    def assert_post(self, uri, *args, **kw):
        return self.assert_request(uri, 'post', *args, **kw)

    def assert_put(self, uri, *args, **kw):
        return self.assert_request(uri, 'put', *args, **kw)

    def assert_delete(self, uri, *args, **kw):
        return self.assert_request(uri, 'delete', *args, **kw)


def encode_multipart_data(fields=None, files=None):  # pragma: no cover
    boundary = ''.join(['-----', base64.urlsafe_b64encode(urandom(27)).decode()])
    crlf = b'\r\n'
    lines = []

    if fields:
        for key, value in fields.items():
            lines.append('--' + boundary)
            lines.append('Content-Disposition: form-data; name="%s"' % key)
            lines.append('')
            lines.append(value if isinstance(value, str) else str(value))

    if files:
        for key, file_path in files.items():
            filename = path.split(file_path)[1]
            lines.append('--' + boundary)
            lines.append(
                'Content-Disposition: form-data; name="%s"; filename="%s"' %
                (key, filename))
            lines.append(
                'Content-Type: %s' %
                (mimetypes.guess_type(filename)[0] or 'application/octet-stream'))
            lines.append('')
            with open(file_path, 'rb') as f:
                lines.append(f.read())

    lines.append('--' + boundary + '--')
    lines.append('')

    body = io.BytesIO()
    length = 0
    for l in lines:
        # noinspection PyTypeChecker
        line = (l if isinstance(l, bytes) else l.encode()) + crlf
        length += len(line)
        body.write(line)
    body.seek(0)
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return content_type, body, length


def find_free_tcp_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind((socket.gethostname(), 0))
        return s.getsockname()[1]
    finally:
        s.close()


if __name__ == '__main__':  # pragma: no cover
    ct, b, len_ = encode_multipart_data(dict(test1='TEST1VALUE'), files=dict(cat='stuff/cat.jpg'))
    print(ct)
    print(len_)
    print(b.read())




import sys
import logging

import ujson

from nanohttp.contexts import Context, context
from nanohttp.exceptions import HttpStatus
from nanohttp.configuration import settings


logger = logging.getLogger('nanohttp')


class Application:
    __logger__ = logger
    __root__ = None

    def __init__(self, root=None):
        self.__root__ = root
        self._hook('app_init')

    def _hook(self, name, *args, **kwargs):
        if hasattr(self, name):
            return getattr(self, name)(*args, **kwargs)

    def _handle_exception(self, ex):
        """This method should return a tuple of (status, resp_generator) and or raise the exception.

        :param ex: The exception to examine
        :return: status, resp_generator
        """
        if isinstance(ex, HttpStatus):
            return ex.status, ex.render()
        raise ex

    def __call__(self, environ, start_response):
        ctx = Context(environ, self)
        ctx.__enter__()
        # start_response("200 OK", [('Content-Type', 'text/plain; charset=utf-8')])

        status = '200 OK'
        buffer = None

        try:
            self._hook('begin_request')
            if self.__root__.__remove_trailing_slash__:
                ctx.path = ctx.path.rstrip('/')

            result = self.__root__(*ctx.path.split('?')[0][1:].split('/'))
            if result:
                resp_generator = iter(result)
                buffer = next(resp_generator)
            else:
                resp_generator = None

        except Exception as ex:
            self.__logger__.exception('Exception while handling the request.')
            status, resp_generator = self._handle_exception(ex)

        self._hook('begin_response')

        if context.response_cookies:
            for cookie in context.response_cookies:
                ctx.response_headers.add_header(*cookie.http_set_cookie())
        start_response(status, ctx.response_headers.items())

        def _response():
            try:
                if buffer is not None:
                    yield ctx.encode_response(buffer)

                if resp_generator:
                    # noinspection PyTypeChecker
                    for chunk in resp_generator:
                        yield ctx.encode_response(chunk)
                else:
                    yield b''
            except Exception as ex_:  # pragma: no cover
                self.__logger__.exception('Exception while serving the response.')
                if settings.debug:
                    yield str(ex_).encode()
                raise ex_

            finally:
                self._hook('end_response')
                context.__exit__(*sys.exc_info())

        return _response()





from .exceptions import HttpStatus, HttpBadRequest, HttpUnauthorized, HttpForbidden, HttpNotFound, \
    HttpMethodNotAllowed, HttpConflict, HttpGone, HttpRedirect, HttpMovedPermanently, HttpFound, \
    HttpInternalServerError
from .cookies import HttpCookie
from .controllers import Controller, RestController, Static
from .decorators import action, html, json, xml, binary, text
from .helpers import quickstart, LazyAttribute
from .cli import main
from .contexts import context, ContextIsNotInitializedError
from .configuration import settings, configure
from .application import Application

__version__ = '0.8.4'




import unittest
from http import cookies

from nanohttp import Controller, html, context, HttpCookie
from nanohttp.tests.helpers import WsgiAppTestCase


class HttpCookieTestCase(WsgiAppTestCase):

    class Root(Controller):

        @html
        def index(self):
            counter = int(context.cookies.get('test-cookie', 0))
            counter += 1
            context.response_cookies.append(HttpCookie('test-cookie', value=str(counter), max_age=1))
            context.response_cookies.append(HttpCookie('dummy-cookie', value=str(counter)))
            context.response_cookies.append(HttpCookie('dummy-cookie3', value=str(counter), domain='example.com'))
            yield 'Index'

        @html
        def secure(self):
            context.response_cookies.append(HttpCookie('dummy-cookie1', value='dummy', http_only=True))
            context.response_cookies.append(HttpCookie('dummy-cookie2', value='dummy', domain='example.com'))
            context.response_cookies.append(HttpCookie('dummy-cookie3', value='dummy', secure=True))
            yield 'Secure'

        @html
        def clear(self):
            context.response_cookies.append(HttpCookie.delete('dummy-cookie'))
            yield 'remove'

    def test_cookie(self):

        response, content = self.assert_get('/')
        cookies_ = cookies.SimpleCookie(response['set-cookie'])
        self.assertEqual(cookies_['test-cookie'].value, '1')
        self.assertIn('dummy-cookie', cookies_)

        response, content = self.assert_get('/', cookies=cookies_)
        cookies_ = cookies.SimpleCookie(response['set-cookie'])
        self.assertEqual(cookies_['test-cookie'].value, '2')
        self.assertIn('dummy-cookie', cookies_)

        response, content = self.assert_get('/secure')
        cookies_ = cookies.SimpleCookie(response['set-cookie'])
        self.assertNotIn('dummy-cookie1', cookies_)
        self.assertNotIn('dummy-cookie2', cookies_)
        self.assertNotIn('dummy-cookie3', cookies_)

        response, content = self.assert_get('/clear')
        cookies_ = cookies.SimpleCookie(response['set-cookie'])
        self.assertIn('dummy-cookie', cookies_)
        self.assertEqual(cookies_['dummy-cookie']['expires'], 'Sat, 01 Jan 2000 00:00:01 GMT')

    def test_cookie_errors(self):
        self.assert_get('/', cookies='Invalid; token; data', status=400)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import pymlconf


settings = pymlconf.DeferredConfigManager()


BUILTIN_CONFIGURATION = """
debug: true
domain:
cookie:
  http_only: false
  secure: false
json:
  indent: 4
"""


def configure(*args, **kwargs):
    settings.load(*args, builtin=BUILTIN_CONFIGURATION, **kwargs)




import re
import unittest

from nanohttp import Controller, text, ContextIsNotInitializedError, context, Application
from nanohttp.contexts import Context
from nanohttp.tests.helpers import WsgiAppTestCase


class ContextProxyTestCase(unittest.TestCase):

    def test_context(self):
        self.assertRaises(ContextIsNotInitializedError, Context.get_current)


class ContextTestCase(WsgiAppTestCase):

    class Root(Controller):

        @text
        def get_uri(self):
            yield context.request_uri

        @text
        def get_scheme(self):
            yield context.request_scheme

        @text('post')
        def get_request_content_length(self):
            yield str(context.request_content_length)

        @text
        def example(self):
            return 'example' if context.example else ''

    class Application(Application):
        @staticmethod
        def begin_request():
            context.example = True

        @staticmethod
        def end_response():
            del context.example

    def test_redirect_response_header(self):
        self.assert_get(
            '/get_uri',
            query_string={'a': 1, 'b': 2},
            expected_response=re.compile('http://nanohttp.org/get_uri\?[ab=12&]+')
        )

        self.assert_get(
            '/get_scheme',
            expected_response='http'
        )

        self.assert_post(
            '/get_request_content_length',
            fields={
                'a': 1,
                'b': [2, 3],
            },
            expected_response='18'
        )

        self.assert_get(
            '/example',
            expected_response='example'
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import sys
from os import chdir
from os.path import relpath, basename

import nanohttp
from .helpers import quickstart, load_controller_from_file
from .configuration import configure

DEFAULT_CONFIG_FILE = 'nanohttp.yml'
DEFAULT_ADDRESS = '8080'


def parse_arguments(argv=None):
    import argparse

    argv = argv or sys.argv

    parser = argparse.ArgumentParser(prog=basename(argv[0]))
    parser.add_argument('-c', '--config-file', action='append', default=[], help='This option may be passed multiple '
                                                                                 'times.')
    parser.add_argument('-d', '--config-directory', action='append', default=[], help='This option may be passed '
                                                                                      'multiple times.')
    parser.add_argument('-b', '--bind', default=DEFAULT_ADDRESS, metavar='{HOST:}PORT', help='Bind Address. default: '
                                                                                             '%s' % DEFAULT_ADDRESS)
    parser.add_argument('-C', '--directory', default='.', help='Change to this path before starting the server '
                                                               'default is: `.`')
    parser.add_argument('-V', '--version', default=False, action='store_true', help='Show the version.')
    parser.add_argument('controller', nargs='?', metavar='{MODULE{.py}}{:CLASS}',
                        help='The python module and controller class to launch. default is python built-in\'s : '
                             '`demo_app`, And the default value for `:CLASS` is `:Root` if omitted.')

    return parser.parse_args(argv[1:])


def main(argv=None):
    args = parse_arguments(argv=argv)

    if args.version:  # pragma: no cover
        print(nanohttp.__version__)
        return 0

    try:
        host, port = args.bind.split(':') if ':' in args.bind else ('', args.bind)

        # Change dir
        if relpath(args.directory, '.') != '.':
            chdir(args.directory)

        configure(files=args.config_file, dirs=args.config_directory, force=True)

        quickstart(
            controller=load_controller_from_file(args.controller),
            host=host,
            port=int(port)
        )
    except KeyboardInterrupt:  # pragma: no cover
        print('CTRL+C detected.')
        return -1
    else:  # pragma: no cover
        return 0




import time
from datetime import datetime

from .configuration import settings
from .constants import HTTP_DATETIME_FORMAT


class HttpCookie(object):
    """
    HTTP Cookie
    http://www.ietf.org/rfc/rfc2109.txt

    ``domain``, ``secure`` and ``http_only`` are taken from ``config`` if not set.
    """
    __slots__ = ('name', 'value', 'path', 'expires',
                 'domain', 'secure', 'http_only')

    def __init__(self, name, value=None, path='/', expires=None, max_age=None, domain=None, secure=None,
                 http_only=None):
        self.name = name
        self.value = value
        self.path = path
        if max_age is None:
            self.expires = expires
        else:
            self.expires = datetime.utcfromtimestamp(time.time() + max_age)
        if domain is None:
            self.domain = settings.domain
        else:
            self.domain = domain
        if secure is None:
            self.secure = settings.cookie.secure
        else:
            self.secure = secure
        if http_only is None:
            self.http_only = settings.cookie.http_only
        else:
            self.http_only = http_only

    @classmethod
    def delete(cls, name, **kwargs):
        """ Returns a cookie to be deleted by browser.
        """
        return cls(name, expires='Sat, 01 Jan 2000 00:00:01 GMT', **kwargs)

    def http_set_cookie(self):
        """ Returns Set-Cookie response header.
        """
        directives = []
        append = directives.append
        append(self.name + '=')
        if self.value:
            append(self.value)
        if self.domain:
            append('; domain=%s' % self.domain)
        if self.expires:
            append('; expires=%s' % (
                self.expires if isinstance(self.expires, str)
                else self.expires.strftime(HTTP_DATETIME_FORMAT))
            )
        if self.path:
            append('; path=%s' % self.path)
        if self.secure:
            append('; secure')
        if self.http_only:
            append('; httponly')
        return 'Set-Cookie', ''.join(directives)
